/*
 * (C) 2007-2012 Alibaba Group Holding Limited.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * Authors:
 *   wuhua <wq163@163.com> , boyan <killme2008@gmail.com>
 */
package com.taobao.metamorphosis.server.transaction.store;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.concurrent.atomic.AtomicInteger;


/**
 * һļ
 * 
 * @author dogun (yuexuqiang at gmail.com)
 */
class DataFile {
    private final int number;
    private final File file;
    private final AtomicInteger referenceCount = new AtomicInteger(0);
    protected FileChannel fc;
    protected RandomAccessFile raf;


    /**
     * 캯ָļҽָָļβ
     * 
     * @param file
     * @throws IOException
     */
    DataFile(final File file) throws IOException {
        this(file, 0, false);
    }


    DataFile(final File file, final int number) throws IOException {
        this(file, number, false);
    }


    int getNumber() {
        return this.number;
    }


    /**
     * 캯ָļҽָָļβ
     * 
     * @param file
     * @throws IOException
     */
    DataFile(final File file, final int number, final boolean force) throws IOException {
        this.file = file;
        this.number = number;
        this.raf = new RandomAccessFile(file, force ? "rws" : "rw");
        this.fc = this.raf.getChannel();
        // ָƵ
        this.fc.position(this.fc.size());
    }


    /**
     * ļĴС
     * 
     * @return ļĴС
     * @throws IOException
     */
    long getLength() throws IOException {
        return this.fc.size();
    }


    void truncate(final long size) throws IOException {
        this.fc.truncate(size);
    }


    /**
     * ȡļ޸ʱ
     * 
     * @return
     * @throws IOException
     */
    long lastModified() throws IOException {
        return this.file.lastModified();
    }


    /**
     * ɾļ
     * 
     * @return Ƿɾɹ
     * @throws IOException
     */
    boolean delete() throws IOException {
        this.close();
        return this.file.delete();
    }


    /**
     * ǿƽдӲ
     * 
     * @throws IOException
     */
    void force() throws IOException {
        if (this.fc.isOpen()) {
            this.fc.force(false);
        }
    }


    /**
     * رļ
     * 
     * @throws IOException
     */
    void close() throws IOException {
        try {
            if (this.fc != null) {
                this.fc.close();
            }
        }
        finally {
            if (this.raf != null) {
                this.raf.close();
            }
        }
    }


    /**
     * ļȡݵbfֱ߶ļβ <br />
     * ļָƶbfĴС
     * 
     * @param bf
     * @throws IOException
     */
    void read(final ByteBuffer bf) throws IOException {
        while (bf.hasRemaining()) {
            final int l = this.fc.read(bf);
            if (l < 0) {
                break;
            }
        }
    }


    /**
     * ļƶλöȡݵbfֱ߶ļβ <br />
     * ļָ벻ƶ
     * 
     * @param bf
     * @param offset
     * @throws IOException
     */
    void read(final ByteBuffer bf, final long offset) throws IOException {
        int size = 0;
        while (bf.hasRemaining()) {
            final int l = this.fc.read(bf, offset + size);
            size += l;
            if (l < 0) {
                break;
            }
        }
    }


    /**
     * дbfȵݵļļָƶ
     * 
     * @param bf
     * @return дļposition
     * @throws IOException
     */
    long write(final ByteBuffer bf) throws IOException {
        while (bf.hasRemaining()) {
            final int l = this.fc.write(bf);
            if (l < 0) {
                break;
            }
        }
        return this.fc.position();
    }


    /**
     * ָλдbfȵݵļļָ<b></b>ƶ
     * 
     * @param offset
     * @param bf
     * @throws IOException
     */
    void write(final long offset, final ByteBuffer bf) throws IOException {
        int size = 0;
        while (bf.hasRemaining()) {
            final int l = this.fc.write(bf, offset + size);
            size += l;
            if (l < 0) {
                break;
            }
        }
    }


    /**
     * ļһü
     * 
     * @return Ӻü
     */
    int increment() {
        return this.referenceCount.incrementAndGet();
    }


    /**
     * ļ
     * 
     * @param i
     */
    void setReferenceCount(final int i) {
        this.referenceCount.set(i);
    }


    /**
     * ļһü
     * 
     * @return ٺü
     */
    int decrement() {
        return this.referenceCount.decrementAndGet();
    }


    /**
     * ļǷʹãüǷ0ˣ
     * 
     * @return ļǷʹ
     */
    boolean isUnUsed() {
        return this.getReferenceCount() <= 0;
    }


    long position() throws IOException {
        return this.fc.position();
    }


    /**
     * üֵ
     * 
     * @return üֵ
     */
    int getReferenceCount() {
        return this.referenceCount.get();
    }


    @Override
    public String toString() {
        String result = null;
        try {
            result =
                    this.file.getName() + " , length = " + this.getLength() + " refCount = " + this.referenceCount
                            + " position:" + this.fc.position();
        }
        catch (final IOException e) {
            result = e.getMessage();
        }
        return result;
    }
}